#include "snapshotcanvas.h"

#include <QDebug>
#include <QGraphicsScene>
#include <QGraphicsItem>
#include <QPixmap>
#include <QPointF>
#include <QMouseEvent>
#include <QList>
#include <QSize>
#include <QApplication>
#include <QDesktopWidget>
#include <QScrollBar>
#include <QStyle>
#include <QResizeEvent>
#include <typeinfo>

#include "items/arrowgraphicsitem.h"
#include "items/boxgraphicsitem.h"
#include "items/boxtextgraphicsitem.h"
#include "items/ellipsegraphicsitem.h"

SnapshotCanvas::SnapshotCanvas(QWidget *parent)
    : QGraphicsView(parent),
      m_origin(0, 0),
      m_itemToDraw(SnapshotCanvas::Arrow),
      m_currentDrawingItem(0),
      m_scene(0),
      m_mouseDown(false),
      m_pixmapItem(0),
      m_pixmapDropShadowEffect(new QGraphicsDropShadowEffect(this)),
      m_captureKeyboardEvents(true)
{
    m_pixmapDropShadowEffect->setColor(QColor(63, 63, 63, 220));
    m_pixmapDropShadowEffect->setOffset(0, 0);
    m_pixmapDropShadowEffect->setBlurRadius(20);
}

SnapshotCanvas::~SnapshotCanvas()
{
    for(int i = 0; i<m_itemsClipboard.size(); ++i) {
        delete m_itemsClipboard.at(i);
    }
}

void SnapshotCanvas::setPixmap(const QPixmap &pixmap)
{
    delete m_scene;
    m_scene = new QGraphicsScene(this);

    connect(m_scene, SIGNAL(selectionChanged()),
            this, SLOT(slotSelectionChanged()));

    m_scene->setBackgroundBrush(QColor(180, 180, 180));

    m_pixmapItem = new QGraphicsPixmapItem(pixmap);
    m_pixmapItem->setGraphicsEffect(m_pixmapDropShadowEffect);

    m_scene->addItem(m_pixmapItem);
    m_scene->setSceneRect(pixmap.rect());

    setScene(m_scene);

    adjustMaximumSize();
}

QFont SnapshotCanvas::textItemFont() const
{
    return m_font;
}

void SnapshotCanvas::setItemColor(const QColor &color)
{
    if (m_color == color) {
        return;
    }

    m_color = color;

    if (m_scene) {
        QList<QGraphicsItem *> items = m_scene->selectedItems();
        for (int i=0; i<items.length(); i++) {
            static_cast<KaptionGraphicsItem*>(items.at(i))->setShapeColor(color);
        }
    }
}

void SnapshotCanvas::setItemScale(const Scale &scale)
{
    if (scale == m_itemScale) {
        return;
    }

    m_itemScale = scale;

    if (m_scene) {
        QList<QGraphicsItem *> items = m_scene->selectedItems();
        for (int i=0; i<items.length(); i++) {
            static_cast<KaptionGraphicsItem*>(items.at(i))->setShapeScale(scale);
        }
    }
}

void SnapshotCanvas::setTextItemFont(const QFont &font)
{
    if (font == m_font) {
        return;
    }

    m_font = font;

    if (m_scene) {
        QList<QGraphicsItem *> items = m_scene->selectedItems();
        for (int i=0; i<items.length(); i++) {
            BoxTextGraphicsItem *item = dynamic_cast<BoxTextGraphicsItem*>(items.at(i));
            if (item) {
                item->setFont(font);
            }
        }
    }
}

void SnapshotCanvas::setItemToDraw(SnapshotCanvas::GraphicItem item)
{
    m_itemToDraw = item;
}

void SnapshotCanvas::deselectItems()
{
    m_scene->clearSelection();
}

void SnapshotCanvas::resizeEvent(QResizeEvent *event)
{
    adjustMaximumSize();
    QGraphicsView::resizeEvent(event);
}

void SnapshotCanvas::slotSelectionChanged()
{
    if (m_scene != 0 && !m_scene->selectedItems().isEmpty()) {
        m_captureKeyboardEvents = true;
    }
}

void SnapshotCanvas::slotDisableKeyboardEventsCapture()
{
    m_captureKeyboardEvents = false;
}

void SnapshotCanvas::mousePressEvent(QMouseEvent *event)
{
    QGraphicsItem *item = itemAt(event->pos());
    if (item == m_pixmapItem) {
        deselectItems();
        if (event->button() == Qt::LeftButton) {
            m_mouseDown = true;
            m_origin = mapToScene(event->pos());
        }
    } else {
        QGraphicsView::mousePressEvent(event);
    }
}

void SnapshotCanvas::mouseMoveEvent(QMouseEvent *event)
{
    QPointF pos = mapToScene(event->pos());

    if (m_mouseDown) {
        if (m_currentDrawingItem == 0) {
            addItem();
            m_currentDrawingItem->setShapeColor(m_color);
        }
        m_currentDrawingItem->createShape(m_origin, pos, m_itemScale);
    } else {
        QGraphicsView::mouseMoveEvent(event);
    }
}

void SnapshotCanvas::mouseReleaseEvent(QMouseEvent *event)
{
    m_mouseDown = false;
    m_currentDrawingItem = 0;
    m_origin = QPointF(0, 0);
    QGraphicsView::mouseReleaseEvent(event);
}

void SnapshotCanvas::keyPressEvent(QKeyEvent *event)
{
    if (!m_captureKeyboardEvents) {
        QGraphicsView::keyPressEvent(event);
        return;
    }

    switch(event->key()) {
    case Qt::Key_Backspace:
    case Qt::Key_Delete:
        removeSelectedItems();
        break;
    case Qt::Key_Escape:
        deselectItems();
        break;
    case Qt::Key_Up:
        moveSelectedItemsBy(0, -2);
        break;
    case Qt::Key_Right:
        moveSelectedItemsBy(2, 0);
        break;
    case Qt::Key_Down:
        moveSelectedItemsBy(0, 2);
        break;
    case Qt::Key_Left:
        moveSelectedItemsBy(-2, 0);
        break;
    }

    if (event->matches(QKeySequence::SelectAll)) {
        QPainterPath selectionArea;
        selectionArea.addRect(sceneRect());
        m_scene->setSelectionArea(selectionArea);
    } else if (event->matches(QKeySequence::Copy)) {
        QList<QGraphicsItem *> items = m_scene->selectedItems();
        if (!items.isEmpty()) {
            for(int i = 0; i<m_itemsClipboard.size(); ++i) {
                delete m_itemsClipboard.at(i);
            }
            m_itemsClipboard.clear();
            for (int i = 0; i<items.size(); ++i) {
                m_itemsClipboard << static_cast<KaptionGraphicsItem*>(items.at(i))->copy();
            }
        }
    } else if (event->matches(QKeySequence::Paste) && !m_itemsClipboard.isEmpty()) {
        QList<KaptionGraphicsItem*> tmp;
        for(int i=0; i<m_itemsClipboard.size(); ++i) {
            m_scene->addItem(m_itemsClipboard.at(i));
            tryToConnectTextItem(m_itemsClipboard.at(i));
            QPointF pos = m_itemsClipboard.at(i)->scenePos();
            m_itemsClipboard.at(i)->moveBy(-10, 10);
            tmp << m_itemsClipboard.at(i)->copy();
        }
        m_itemsClipboard.clear();
        m_itemsClipboard = tmp; //recursive pasting
    }
}

void SnapshotCanvas::addItem()
{
    if (m_scene == 0) {
        return;
    }

    switch(m_itemToDraw) {
    case Arrow:
        m_currentDrawingItem = new ArrowGraphicsItem();
        break;
    case Box:
        m_currentDrawingItem = new BoxGraphicsItem();
        break;
    case Text:
        m_currentDrawingItem = new BoxTextGraphicsItem(m_font);
        tryToConnectTextItem(m_currentDrawingItem);
        break;
    case Ellipse:
        m_currentDrawingItem = new EllipseGraphicsItem();
        break;
    }

    if (m_currentDrawingItem != 0) {
        m_scene->addItem(m_currentDrawingItem);
    }
}

void SnapshotCanvas::removeSelectedItems()
{
    QList<QGraphicsItem *> items = m_scene->selectedItems();
    for (int i=0; i<items.length(); i++) {
        //m_scene->removeItem(items.at(i)); <- BSP Tree index QT bug??? Should ask on IRC...
        delete items.at(i);
    }
}

void SnapshotCanvas::moveSelectedItemsBy(int dx, int dy)
{
    QList<QGraphicsItem *> items = m_scene->selectedItems();
    for (int i=0; i<items.length(); i++) {
        items.at(i)->moveBy(dx, dy);
    }
}

void SnapshotCanvas::adjustMaximumSize()
{
    QRectF sr = sceneRect();

    int width = sr.width() + frameWidth() * 2;
    int height = sr.height() + frameWidth() * 2;

    if (minimumWidth() > width) {
        width = minimumWidth();
    }

    if (minimumHeight() > height) {
        height = minimumHeight();
    }

    int scrollBarExtent = style()->pixelMetric(QStyle::PM_ScrollBarExtent);

    if (verticalScrollBar()->isVisible()) {
        width += scrollBarExtent + frameWidth() * 2;
    }

    if (horizontalScrollBar()->isVisible()) {
        height += scrollBarExtent + frameWidth() * 2;
    }

    setMaximumSize(width, height);
}

void SnapshotCanvas::tryToConnectTextItem(KaptionGraphicsItem *item)
{
    BoxTextGraphicsItem *textItem = dynamic_cast<BoxTextGraphicsItem*>(item);
    if (textItem) {
        connect(textItem, SIGNAL(itemDeselected()),
                this, SLOT(deselectItems()));
        connect(textItem, SIGNAL(itemDeselected()),
                this, SLOT(slotDisableKeyboardEventsCapture()));
    }
}
