#include "snapshotpreview.h"
#include "ui_snapshotpreview.h"

// Qt headers
#include <QtGui/QWidget>
#include <QGridLayout>
#include <QLabel>
#include <QCloseEvent>
#include <KImageIO>
#include <QPointer>
#include <KFileDialog>
#include <KUrl>
#include <KLocale>
#include <kio/fileundomanager.h>
#include <KIO/DeleteJob>
#include <KIO/CopyJob>
#include <KIO/JobUiDelegate>
#include <KIO/NetAccess>
#include <KMessageBox>
#include <QByteArray>
#include <KMimeType>
#include <QFile>
#include <QIODevice>
#include <QImageWriter>
#include <KTemporaryFile>
#include <QGraphicsView>
#include <QGraphicsScene>
#include <QPainter>
#include <QSignalMapper>
#include <KMessageBox>
#include <QClipboard>
#include <KProgressDialog>
#include <QStyle>
#include <QSizePolicy>
#include <KPushButton>
#include <KNotification>
#include <KIconLoader>
#include <KFontDialog>
#include "scale.h"
#include "kaptionapplication.h"
#include "graphic/snapshotcanvas.h"
#include "graphic/items/boxtextgraphicsitem.h"
#include "kaptiongraphicsitem.h"
#include "settings.h"
#include "configdialog/configdialog.h"
#include "uploadprogressdialog.h"

SnapshotPreview::SnapshotPreview(QWidget *parent)
    : QFrame(parent)
{
    init();
}

SnapshotPreview::SnapshotPreview(const QPixmap &pixmap, QWidget *parent)
    : QFrame(parent)
{
    init();
    setPixmap(pixmap);
}

SnapshotPreview::~SnapshotPreview()
{
    delete ui;
    if (m_tmpFile) {
        m_tmpFile->deleteLater();
    }
}

void SnapshotPreview::init()
{
    m_tmpFile = 0;
    m_progressDialog = 0;

    ui = new Ui::SnapshotPreview;
    ui->setupUi(this);

    KIconLoader *iconLoader = KIconLoader::global();

    ui->arrowBtn->setIcon(iconLoader->loadIcon("toolbox_arrow", KIconLoader::Small));
    ui->boxBtn->setIcon(iconLoader->loadIcon("toolbox_box", KIconLoader::Small));
    ui->ellipseBtn->setIcon(iconLoader->loadIcon("toolbox_ellipse", KIconLoader::Small));
    ui->textBtn->setIcon(iconLoader->loadIcon("toolbox_text", KIconLoader::Small));

    ui->formatTextBtn->setIcon(iconLoader->loadIcon("draw-text", KIconLoader::Small));

    ui->snapshotCanvas->setItemColor(ui->colorBtn->color());
    slotSetItemScale(ui->widthSlider->sliderPosition());
    QFont font;
    font.setBold(true);
    font.setPointSize(10);
    ui->snapshotCanvas->setTextItemFont(font);

    ui->cancelBtn->setIcon(iconLoader->loadIcon("dialog-cancel", KIconLoader::Small));
    ui->saveBtn->setIcon(iconLoader->loadIcon("document-save-as", KIconLoader::Small));

    m_signalMapper = new QSignalMapper(this);
    m_signalMapper->setMapping(ui->arrowBtn, ARROW);
    m_signalMapper->setMapping(ui->boxBtn, BOX);
    m_signalMapper->setMapping(ui->textBtn, TEXT);
    m_signalMapper->setMapping(ui->ellipseBtn, ELLIPSE);

    connect(ui->arrowBtn, SIGNAL(clicked()),
            m_signalMapper, SLOT(map()));
    connect(ui->boxBtn, SIGNAL(clicked()),
            m_signalMapper, SLOT(map()));
    connect(ui->textBtn, SIGNAL(clicked()),
            m_signalMapper, SLOT(map()));
    connect(ui->ellipseBtn, SIGNAL(clicked()),
            m_signalMapper, SLOT(map()));

    connect(m_signalMapper, SIGNAL(mapped(int)),
            this, SLOT(setItemToDraw(int)));

    connect(ui->colorBtn, SIGNAL(changed(QColor)),
            ui->snapshotCanvas, SLOT(setItemColor(QColor)));

    connect(ui->formatTextBtn, SIGNAL(clicked()),
            this, SLOT(slotOpenFontDlg()));

    connect(ui->widthSlider, SIGNAL(valueChanged(int)),
            this, SLOT(slotSetItemScale(int)));

    connect(ui->cancelBtn, SIGNAL(clicked()),
            this, SLOT(close()));
    connect(ui->saveBtn, SIGNAL(clicked()),
            this, SLOT(slotSaveAs()));
    connect(ui->uploadBtn, SIGNAL(clicked()),
            this, SLOT(slotUpload()));
}

void SnapshotPreview::setPixmap(const QPixmap &pixmap)
{
    ui->snapshotCanvas->setPixmap(pixmap);
    setToolbarsMaxSize();
    setItemToDraw(ARROW);

    connect(ui->snapshotCanvas->scene(), SIGNAL(selectionChanged()),
            this, SLOT(slotUpdateEditingOption()));
}

void SnapshotPreview::slotUpdateEditingOption()
{
    QGraphicsScene *scene = ui->snapshotCanvas->scene();
    QList<QGraphicsItem*> items = scene->selectedItems();
    if (!items.isEmpty()) {
        KaptionGraphicsItem *item = static_cast<KaptionGraphicsItem*>(items.last());

        ui->colorBtn->blockSignals(true);
        ui->widthSlider->blockSignals(true);

        ui->colorBtn->setColor(item->shapeColor());

        int max = ui->widthSlider->maximum();
        int min = ui->widthSlider->minimum();
        ui->widthSlider->setSliderPosition(item->shapeScale().value(min, max));

        ui->colorBtn->blockSignals(false);
        ui->widthSlider->blockSignals(false);
    }
}

void SnapshotPreview::slotSetItemScale(int width)
{
    Scale scale(ui->widthSlider->minimum(), ui->widthSlider->maximum());
    scale.setValue(width);
    ui->snapshotCanvas->setItemScale(scale);
}

void SnapshotPreview::slotOpenFontDlg()
{
    QGraphicsScene *scene = ui->snapshotCanvas->scene();
    QList<QGraphicsItem*> items = scene->selectedItems();
    QFont font = ui->snapshotCanvas->textItemFont();
    if (!items.isEmpty()) {
        for(int i=0; i<items.size(); ++i) {
            BoxTextGraphicsItem *item = dynamic_cast<BoxTextGraphicsItem*>(items.at(i));
            if (item) {
                font = item->font();
                break;
            }
        }
    }

    int result = KFontDialog::getFont(font);
    if (result == KFontDialog::Accepted) {
        ui->snapshotCanvas->setTextItemFont(font);
    }
}

void SnapshotPreview::setItemToDraw(int i)
{
    //FIXME: this is real shit...
    switch(i) {
    case ARROW:
        ui->snapshotCanvas->setItemToDraw(SnapshotCanvas::Arrow);
        ui->arrowBtn->setChecked(true);
        ui->arrowBtn->setDisabled(true);
        ui->boxBtn->setChecked(false);
        ui->boxBtn->setDisabled(false);
        ui->textBtn->setChecked(false);
        ui->textBtn->setDisabled(false);
        ui->ellipseBtn->setChecked(false);
        ui->ellipseBtn->setDisabled(false);
        break;
    case BOX:
        ui->snapshotCanvas->setItemToDraw(SnapshotCanvas::Box);
        ui->arrowBtn->setChecked(false);
        ui->arrowBtn->setDisabled(false);
        ui->boxBtn->setChecked(true);
        ui->boxBtn->setDisabled(true);
        ui->textBtn->setChecked(false);
        ui->textBtn->setDisabled(false);
        ui->ellipseBtn->setChecked(false);
        ui->ellipseBtn->setDisabled(false);
        break;
    case TEXT:
        ui->snapshotCanvas->setItemToDraw(SnapshotCanvas::Text);
        ui->arrowBtn->setChecked(false);
        ui->arrowBtn->setDisabled(false);
        ui->boxBtn->setChecked(false);
        ui->boxBtn->setDisabled(false);
        ui->textBtn->setChecked(true);
        ui->textBtn->setDisabled(true);
        ui->ellipseBtn->setChecked(false);
        ui->ellipseBtn->setDisabled(false);
        break;
    case ELLIPSE:
        ui->snapshotCanvas->setItemToDraw(SnapshotCanvas::Ellipse);
        ui->arrowBtn->setChecked(false);
        ui->arrowBtn->setDisabled(false);
        ui->boxBtn->setChecked(false);
        ui->boxBtn->setDisabled(false);
        ui->textBtn->setChecked(false);
        ui->textBtn->setDisabled(false);
        ui->ellipseBtn->setChecked(true);
        ui->ellipseBtn->setDisabled(true);
        break;
    }
}

void SnapshotPreview::slotSaveAs()
{
    QStringList mimetypes = KImageIO::mimeTypes(KImageIO::Writing);

    QString filename = "image_" + QDateTime::currentDateTime().toString("yyyyMMdd-hhmmss") + ".png";
    QPointer<KFileDialog> dlg = new KFileDialog(QDir::homePath().append(QDir::separator()).append(filename), mimetypes.join(" "), this);

    dlg->setOperationMode(KFileDialog::Saving);
    dlg->setCaption(i18n("Save As"));

    if (!dlg->exec()) {
        delete dlg;
        return;
    }

    KUrl url = dlg->selectedUrl();
    if (!url.isValid()) {
        delete dlg;
        return;
    }

    if (KIO::NetAccess::exists(url, KIO::NetAccess::DestinationSide, this)) {
        const QString title = i18n("File Exists");
        const QString text = i18n("<qt>Do you really want to overwrite <b>%1</b>?</qt>", url.prettyUrl());
        if (KMessageBox::Continue != KMessageBox::warningContinueCancel(this, text, title, KGuiItem(i18n("Overwrite")))) {
            delete dlg;
            return;
        }
    }

    QByteArray type = "PNG";
    QString mime = KMimeType::findByUrl(url.fileName(), 0, url.isLocalFile(), true)->name();
    const QStringList types = KImageIO::typeForMime(mime);
    if (!types.isEmpty()) {
        type = types.first().toLatin1();
    }

    bool ok = false;

    if (url.isLocalFile()) {
        QFile output(url.toLocalFile());
        if (output.open(QFile::WriteOnly)) {
            ok = saveImage(&output, type);
        }
    } else {
        KTemporaryFile tmpFile;
        if (tmpFile.open()) {
            if (saveImage(&tmpFile, type)) {
                ok = KIO::NetAccess::upload(tmpFile.fileName(), url, this);
            }
        }
    }

    QApplication::restoreOverrideCursor();
    if (!ok) {
        const QString caption = i18n("Unable to Save Image");
        const QString text = i18n("Kaption was unable to save the image to\n%1.", url.prettyUrl());
        KMessageBox::error(this, text, caption);
    }

    delete dlg;
}

void SnapshotPreview::slotUpload()
{
    Settings *settings = Settings::self();

    QString scheme    = settings->type() == Settings::EnumType::FTP ? "ftp" : "sftp";
    QString server    = settings->server();
    uint    port      = settings->port();
    QString directory = settings->directory();
    QString username  = settings->username();
    QString password  = settings->password();

    if (server.isEmpty() || port < 1) {
        QString txt = i18n("To upload the screenshot you must configure a connection.\n"
                           "Do it now?");
        QString title = i18n("Missing connection configuration");
        if (KMessageBox::warningYesNo(this, txt, title) == KMessageBox::Yes) {
            (static_cast<KaptionApplication*>(kapp))->slotConfigKaption("FtpConfig");
        }
    } else {
        // TODO: All this code should be placed in something like
        //       an upload service
        m_tmpFile =  new KTemporaryFile;
        if (m_tmpFile->open()) {
            if (saveImage(m_tmpFile, "png")) {
                KUrl url;
                url.setScheme(scheme);
                url.setHost(server);
                url.setPort(port);
                url.setPath(directory+"/");
                url.setFileName(QFileInfo(m_tmpFile->fileName()).baseName()+".png");
                if (!username.isNull() && !password.isNull()) {
                    url.setUserName(username);
                    url.setPassword(password);
                }
                // Before uploading I must set permission to 0666
                m_tmpFile->setPermissions(QFile::ReadOwner | QFile::WriteOwner |
                                          QFile::ReadGroup | QFile::WriteGroup |
                                          QFile::ReadOther | QFile::WriteOther);
                m_currentJob = KIO::copy(m_tmpFile->fileName(), url, KIO::HideProgressInfo);
                m_currentJob->setUiDelegate(0);
                connect(m_currentJob, SIGNAL(infoMessage(KJob*,QString)),
                        this, SLOT(slotPrintUploadInfo(KJob*,QString)));
                connect(m_currentJob, SIGNAL(description(KJob*,QString)),
                        this, SLOT(slotPrintUploadInfo(KJob*,QString)));
                connect(m_currentJob, SIGNAL(warning(KJob*,QString)),
                        this, SLOT(slotPrintUploadInfo(KJob*,QString)));
                connect(m_currentJob, SIGNAL(percent(KJob*,unsigned long)),
                        this, SLOT(slotUploading(KJob*,unsigned long)));
                connect(m_currentJob, SIGNAL(result(KJob*)),
                        this, SLOT(slotUploadResult(KJob*)));

                if (m_progressDialog == 0) {
                    m_progressDialog = new UploadProgressDialog(this, i18n("Upload progress"));
                    connect(m_progressDialog, SIGNAL(cancelClicked()),
                            this, SLOT(slotCancelUpload()));
                }
                m_progressDialog->setLabelText(i18n("Uploading to %1", url.prettyUrl()));
                m_progressDialog->clearLogInfo();
                m_progressDialog->setButtons(KDialog::Cancel);
                m_progressDialog->progressBar()->setValue(0);
                m_progressDialog->show();
            }
        }
    }
}

void SnapshotPreview::slotPrintUploadInfo(KJob *job, const QString &plain)
{
    Q_UNUSED(job)
    m_progressDialog->setLogInfo(QIcon::fromTheme("documentinfo"), plain);
}

void SnapshotPreview::slotUploading(KJob *job, unsigned long percent)
{
    Q_UNUSED(job)
    m_progressDialog->progressBar()->setValue(percent);
}

// TODO: Write this function better...it is horrible!
void SnapshotPreview::slotUploadResult(KJob *job)
{
    KIO::CopyJob *copyJob = static_cast<KIO::CopyJob*>(job);
    QString n_eventid;
    QString n_txt;
    QPixmap n_pixmap;
    //QStringList n_actions;

    if (job->error()) {
        QString text;
        if (!job->errorString().isEmpty()) {
            text = job->errorString();
        } else {
            text = i18n("Upload canceled");
        }

        QIcon icon = QIcon::fromTheme("dialog-close");

        // Setup "uploaderror" notification
        n_eventid = "uploaderror";
        n_txt = text;
        n_pixmap = icon.pixmap(64, 64);

        // Show error into the progress dialog
        m_progressDialog->setLogInfo(icon, text);
        m_progressDialog->setButtons(KDialog::Close);

        // Delete remote partial file
        KUrl partial = copyJob->destUrl();
        partial.setFileName(partial.fileName()+".part");
        KIO::del(partial, KIO::HideProgressInfo);
    } else {
        if (m_progressDialog->isVisible()) {
            // Set 100% progress, so it looks pretty
            m_progressDialog->progressBar()->setValue(100);
            m_progressDialog->hide();
        }

        QString txt = i18n("Image uploaded");
        if (Settings::useClipboard() && !Settings::clipboardContents().isEmpty()) {
            QString clipboard = Settings::clipboardContents();
            clipboard.replace("%filename", copyJob->destUrl().fileName());
            kapp->clipboard()->setText(clipboard);
            if (clipboard.startsWith("http://", Qt::CaseInsensitive)) {
                clipboard = QString("<a href=\"%1\">%1</a>").arg(clipboard);
            } else {
                clipboard = QString("<em>%1</em>").arg(clipboard);
            }
            txt += i18n("<br><br>Clipboard content:<br>%1<br>", clipboard);
        }

        // Setup "uploadok" notification
        n_eventid = "uploadok";
        n_txt = txt;
        n_pixmap = getFinalPixmap();
    }

    // Launch notification
    KNotification *notification = new KNotification(n_eventid, this);
    notification->setTitle("Kaption");
    notification->setPixmap(n_pixmap);
    notification->setText(n_txt);
    notification->sendEvent();

    // House cleaning
    m_currentJob->deleteLater();
    m_tmpFile->deleteLater();
}

void SnapshotPreview::slotCancelUpload()
{
    if (m_currentJob != 0 && m_currentJob->percent() < 100) {
        m_currentJob->kill(KJob::EmitResult);
    }
}

bool SnapshotPreview::saveImage(QIODevice *device, const QByteArray &format)
{
    QImageWriter imgWriter(device, format);

    if (!imgWriter.canWrite()) {
        return false;
    }

    if (0 == qstricmp(format.constData(), "jpeg") || 0 == qstricmp(format.constData(), "jpg")) {
        imgWriter.setQuality(100);
    }

    QPixmap p = getFinalPixmap();
    return imgWriter.write(p.toImage());
}

QPixmap SnapshotPreview::getFinalPixmap() const
{
    ui->snapshotCanvas->deselectItems();
    QGraphicsScene *scene = ui->snapshotCanvas->scene();
    QPixmap p(scene->width(), scene->height());
    QPainter painter(&p);
    painter.setRenderHint(QPainter::Antialiasing);
    scene->render(&painter);
    return p;
}

void SnapshotPreview::showEvent(QShowEvent *e)
{
    QFrame::showEvent(e);
}

void SnapshotPreview::resizeEvent(QResizeEvent *e)
{
    setToolbarsMaxSize();
    QFrame::resizeEvent(e);
}

void SnapshotPreview::setToolbarsMaxSize()
{
    int maxHeight = ui->leftToolbar->layout()->minimumSize().height();
    if (ui->rightToolbar->layout()->minimumSize().height() > maxHeight) {
        maxHeight = ui->rightToolbar->layout()->minimumSize().height();
    }
    if (ui->centralLayout->maximumSize().height() > maxHeight) {
        maxHeight = ui->centralLayout->maximumSize().height();
    }
    ui->leftToolbar->setMaximumHeight(maxHeight);
    ui->rightToolbar->setMaximumHeight(maxHeight);

    int maxWidth = ui->bottomToolbar->layout()->minimumSize().width();
    if (ui->snapshotCanvas->maximumWidth() > maxWidth) {
        maxWidth = ui->snapshotCanvas->maximumWidth();
    }
    ui->bottomToolbar->setMaximumWidth(maxWidth);

    layout()->invalidate();
}
